@@ -737,13 +737,13 @@ qdisk_init(char *partname, char *label)
 
 	ret = qdisk_validate(partname);
 	if (ret < 0) {
-		logt_print(LOG_DEBUG, "qdisk_verify");
+		logt_print(LOG_DEBUG, "qdisk_validate: %s\n", strerror(errno));
 		return -1;
 	}
 
 	ret = qdisk_open(partname, &disk);
 	if (ret < 0) {
-		logt_print(LOG_ERR, "qdisk_open");
+		logt_print(LOG_ERR, "qdisk_open: %s\n", strerror(errno));
 		return -1;
 	}
 
