@@ -4,6 +4,7 @@ import sys, getopt, time, os
 import pexpect, re
 import telnetlib
 import atexit
+import __main__
 
 ## do not add code here.
 #BEGIN_VERSION_GENERATION
@@ -12,10 +13,6 @@ REDHAT_COPYRIGHT = ""
 BUILD_DATE = "March, 2008"
 #END_VERSION_GENERATION
 
-POWER_TIMEOUT = 20
-SHELL_TIMEOUT = 3
-LOGIN_TIMEOUT = 5
-
 LOG_MODE_VERBOSE = 100
 LOG_MODE_QUIET = 0
 
@@ -78,6 +75,14 @@ all_opt = {
 		"shortdesc" : "Fencing Action",
 		"default" : "reboot",
 		"order" : 1 },
+	"io_fencing" : {
+		"getopt" : "o:",
+		"longopt" : "action",
+		"help" : "-o, --action=<action>          Action: status, enable or disable",
+		"required" : "1",
+		"shortdesc" : "Fencing Action",
+		"default" : "disable",
+		"order" : 1 },
 	"ipaddr" : {
 		"getopt" : "a:",
 		"longopt" : "ip",
@@ -314,9 +319,46 @@ all_opt = {
 		"default" : ",", 
 		"required" : "0",
 		"shortdesc" : "Separator for CSV created by operation list",
-		"order" : 100 }
+		"order" : 100 },
+	"login_timeout" : {
+		"getopt" : "y:",
+		"longopt" : "login-timeout",
+		"help" : "--login-timeout <seconds>      Wait X seconds for cmd prompt after login",
+		"default" : "5", 
+		"order" : 200 },
+	"shell_timeout" : {
+		"getopt" : "Y:",
+		"longopt" : "shell-timeout",
+		"help" : "--shell-timeout <seconds>      Wait X seconds for cmd prompt after issuing command",
+		"default" : "3", 
+		"order" : 200 },
+	"power_timeout" : {
+		"getopt" : "g:",
+		"longopt" : "power-timeout",
+		"help" : "--power-timeout <seconds>      Test X seconds for status change after ON/OFF",
+		"default" : "20", 
+		"order" : 200 },
+	"power_wait" : {
+		"getopt" : "G:",
+		"longopt" : "power-wait",
+		"help" : "--power-wait <seconds>         Wait X seconds after issuing ON/OFF",
+		"default" : "0", 
+		"order" : 200 },
+	"missing_as_off" : {
+		"getopt" : "M",
+		"longopt" : "missing-as-off",
+		"help" : "--missing-as-off               Missing port returns OFF instead of failure",
+		"order" : 200 },
+	"retry_on" : {
+		"getopt" : "F:",
+		"longopt" : "retry-on",
+		"help" : "--retry-on <attempts>          Count of attempts to retry power on",
+		"default" : "1",
+		"order" : 200 }
 }
 
+common_opt = [ "retry_on" ]
+
 class fspawn(pexpect.spawn):
 	def log_expect(self, options, pattern, timeout):
 		result = self.expect(pattern, timeout)
@@ -330,7 +372,7 @@ def atexit_handler():
 		os.close(1)
 	except IOError:
 		sys.stderr.write("%s failed to close standard output\n"%(sys.argv[0]))
-		sys.exit(1)
+		sys.exit(EC_GENERIC_ERROR)
 
 def version(command, release, build_date, copyright_notice):
 	print command, " ", release, " ", build_date
@@ -341,7 +383,7 @@ def fail_usage(message = ""):
 	if len(message) > 0:
 		sys.stderr.write(message+"\n")
 	sys.stderr.write("Please use '-h' for usage\n")
-	sys.exit(EC_BAD_ARGS)
+	sys.exit(EC_GENERIC_ERROR)
 
 def fail(error_code):
 	message = {
@@ -354,7 +396,7 @@ def fail(error_code):
 		EC_STATUS_HMC : "Failed: Either unable to obtaion correct plug status, partition is not available or incorrect HMC version used"
 	}[error_code] + "\n"
 	sys.stderr.write(message)
-	sys.exit(error_code)
+	sys.exit(EC_GENERIC_ERROR)
 
 def usage(avail_opt):
 	global all_opt
@@ -377,8 +419,10 @@ def metadata(avail_opt, options, docs):
 	sorted_list.sort(lambda x, y: cmp(x[1]["order"], y[1]["order"]))
 
 	print "<?xml version=\"1.0\" ?>"
-	print "<resource-agent name=\"" + os.path.basename(sys.argv[0])[:-3] + "\" shortdesc=\"" + docs["shortdesc"] + "\" >"
+	print "<resource-agent name=\"" + os.path.basename(sys.argv[0]) + "\" shortdesc=\"" + docs["shortdesc"] + "\" >"
 	print "<longdesc>" + docs["longdesc"] + "</longdesc>"
+	if docs.has_key("vendorurl"):
+		print "<vendor-url>" + docs["vendorurl"] + "</vendor-url>"
 	print "<parameters>"
 	for option, value in sorted_list:
 		if all_opt[option].has_key("shortdesc"):
@@ -386,7 +430,7 @@ def metadata(avail_opt, options, docs):
 
 			default = ""
 			if all_opt[option].has_key("default"):
-				default = "default=\""+all_opt[option]["default"]+"\""
+				default = "default=\""+str(all_opt[option]["default"])+"\""
 			elif options.has_key("-" + all_opt[option]["getopt"][:-1]):
 				if options["-" + all_opt[option]["getopt"][:-1]]:
 					try:
@@ -394,7 +438,7 @@ def metadata(avail_opt, options, docs):
 					except TypeError:
 						## @todo/@note: Currently there is no clean way how to handle lists
 						## we can create a string from it but we can't set it on command line
-						default = "default=\"list of values\""
+						default = "default=\"" + str(options["-" + all_opt[option]["getopt"][:-1]]) +"\""
 			elif options.has_key("-" + all_opt[option]["getopt"]):
 				default = "default=\"true\" "
 
@@ -421,12 +465,18 @@ def metadata(avail_opt, options, docs):
 	print "\t<action name=\"status\" />"
 	print "\t<action name=\"list\" />"
 	print "\t<action name=\"monitor\" />"
-	print "\t<action name=\"meta-data\" />"	
+	print "\t<action name=\"metadata\" />"	
 	print "</actions>"
 	print "</resource-agent>"
 
 def process_input(avail_opt):
 	global all_opt
+	global common_opt
+
+	##
+	## Add options which are available for every fence agent
+	#####
+	avail_opt.extend(common_opt)
 
 	##
 	## Set standard environment
@@ -504,11 +554,14 @@ def process_input(avail_opt):
 				name = "ipaddr"
 			elif name == "modulename":
 				name = "module_name"
+			elif name == "action" and 1 == avail_opt.count("io_fencing"):
+				name = "io_fencing"
 
 			##
 			######
 			if avail_opt.count(name) == 0:
-				fail_usage("Parse error: Unknown option '"+line+"'")
+				sys.stderr.write("Parse error: Ignoring unknown option '"+line+"'\n")
+				continue
 
 			if all_opt[name]["getopt"].endswith(":"):
 				opt["-"+all_opt[name]["getopt"].rstrip(":")] = value
@@ -523,7 +576,13 @@ def process_input(avail_opt):
 ######
 def check_input(device_opt, opt):
 	global all_opt
+	global common_opt
 
+	##
+	## Add options which are available for every fence agent
+	#####
+	device_opt.extend(common_opt)
+	
 	options = dict(opt)
 	options["device_opt"] = device_opt
 
@@ -554,9 +613,13 @@ def check_input(device_opt, opt):
 	else:
 		options["log"] = LOG_MODE_QUIET
 
-	if 0 == ["on", "off", "reboot", "status", "list", "monitor"].count(options["-o"].lower()):
-		fail_usage("Failed: Unrecognised action '" + options["-o"] + "'")
-
+	if 0 == device_opt.count("io_fencing"):
+		if 0 == ["on", "off", "reboot", "status", "list", "monitor"].count(options["-o"].lower()):
+			fail_usage("Failed: Unrecognised action '" + options["-o"] + "'")
+	else:
+		if 0 == ["enable", "disable", "status", "list", "monitor"].count(options["-o"].lower()):
+			fail_usage("Failed: Unrecognised action '" + options["-o"] + "'")
+		
 	if (0 == options.has_key("-l")) and device_opt.count("login") and (device_opt.count("no_login") == 0):
 		fail_usage("Failed: You have to set login name")
 
@@ -607,7 +670,7 @@ def check_input(device_opt, opt):
 	return options
 	
 def wait_power_status(tn, options, get_power_fn):
-	for dummy in xrange(POWER_TIMEOUT):
+	for dummy in xrange(int(options["-g"])):
 		if get_power_fn(tn, options) != options["-o"]:
 			time.sleep(1)
 		else:
@@ -633,11 +696,13 @@ def show_docs(options, docs = None):
 		sys.exit(0)
 
 	if options.has_key("-V"):
-		print RELEASE_VERSION, BUILD_DATE
-		print REDHAT_COPYRIGHT
+		print __main__.RELEASE_VERSION, __main__.BUILD_DATE
+		print __main__.REDHAT_COPYRIGHT
 		sys.exit(0)
 
 def fence_action(tn, options, set_power_fn, get_power_fn, get_outlet_list = None):
+	result = 0
+	
 	## Process options that manipulate fencing device
 	#####
 	if (options["-o"] == "list") and (0 == options["device_opt"].count("port")) and (0 == options["device_opt"].count("partition")):
@@ -663,12 +728,25 @@ def fence_action(tn, options, set_power_fn, get_power_fn, get_outlet_list = None
 	if status != "on" and status != "off":  
 		fail(EC_STATUS)
 
+	
+	if options["-o"] == "enable":
+		options["-o"] = "on"
+	if options["-o"] == "disable":
+		options["-o"] = "off"
+
 	if options["-o"] == "on":
 		if status == "on":
 			print "Success: Already ON"
 		else:
-			set_power_fn(tn, options)
-			if wait_power_status(tn, options, get_power_fn):
+			power_on = False
+			for i in range(1,1 + int(options["-F"])):
+				set_power_fn(tn, options)
+				time.sleep(int(options["-G"]))
+				if wait_power_status(tn, options, get_power_fn):
+					power_on = True
+					break
+
+			if power_on:
 				print "Success: Powered ON"
 			else:
 				fail(EC_WAITING_ON)
@@ -677,6 +755,7 @@ def fence_action(tn, options, set_power_fn, get_power_fn, get_outlet_list = None
 			print "Success: Already OFF"
 		else:
 			set_power_fn(tn, options)
+			time.sleep(int(options["-G"]))
 			if wait_power_status(tn, options, get_power_fn):
 				print "Success: Powered OFF"
 			else:
@@ -685,17 +764,32 @@ def fence_action(tn, options, set_power_fn, get_power_fn, get_outlet_list = None
 		if status != "off":
 			options["-o"] = "off"
 			set_power_fn(tn, options)
+			time.sleep(int(options["-G"]))
 			if wait_power_status(tn, options, get_power_fn) == 0:
 				fail(EC_WAITING_OFF)
 		options["-o"] = "on"
-		set_power_fn(tn, options)
-		if wait_power_status(tn, options, get_power_fn) == 0:
+
+		power_on = False
+		for i in range(1,1 + int(options["-F"])):
+			set_power_fn(tn, options)
+			time.sleep(int(options["-G"]))
+			if wait_power_status(tn, options, get_power_fn) == 1:
+				power_on = True
+				break
+
+		if power_on == False:
+			# this should not fail as not was fenced succesfully
 			sys.stderr.write('Timed out waiting to power ON\n')
+
 		print "Success: Rebooted"
 	elif options["-o"] == "status":
 		print "Status: " + status.upper()
+		if status.upper() == "OFF":
+			result = 2
 	elif options["-o"] == "monitor":
 		1
+	
+	return result
 
 def fence_login(options):
 	force_ipvx=""
@@ -712,7 +806,7 @@ def fence_login(options):
 		login_eol = "\r\n"
 
 	try:
-		re_login = re.compile("(login: )|(Login Name:  )|(username: )|(User Name :)", re.IGNORECASE)
+		re_login = re.compile("(login\s*: )|(Login Name:  )|(username: )|(User Name :)", re.IGNORECASE)
 		re_pass  = re.compile("password", re.IGNORECASE)
 
 		if options.has_key("-z"):
@@ -737,21 +831,21 @@ def fence_login(options):
 				
 			if options.has_key("telnet_over_ssh"):
 				#This is for stupid ssh servers (like ALOM) which behave more like telnet (ignore name and display login prompt)
-				result = conn.log_expect(options, [ re_login, "Are you sure you want to continue connecting (yes/no)?" ], LOGIN_TIMEOUT)
+				result = conn.log_expect(options, [ re_login, "Are you sure you want to continue connecting (yes/no)?" ], int(options["-y"]))
 				if result == 1:
 					conn.sendline("yes") # Host identity confirm
-					conn.log_expect(options, re_login, LOGIN_TIMEOUT)
+					conn.log_expect(options, re_login, int(options["-y"]))
 
 				conn.sendline(options["-l"])
-				conn.log_expect(options, re_pass, LOGIN_TIMEOUT)
+				conn.log_expect(options, re_pass, int(options["-y"]))
 			else:
-				result = conn.log_expect(options, [ "ssword:", "Are you sure you want to continue connecting (yes/no)?" ], LOGIN_TIMEOUT)
+				result = conn.log_expect(options, [ "ssword:", "Are you sure you want to continue connecting (yes/no)?" ], int(options["-y"]))
 				if result == 1:
 					conn.sendline("yes")
-					conn.log_expect(options, "ssword:", LOGIN_TIMEOUT)
+					conn.log_expect(options, "ssword:", int(options["-y"]))
 
 			conn.sendline(options["-p"])
-			conn.log_expect(options, options["-c"], LOGIN_TIMEOUT)
+			conn.log_expect(options, options["-c"], int(options["-y"]))
 		elif options.has_key("-x") and 1 == options.has_key("-k"):
 			command = '%s %s %s@%s -i %s -p %s' % (SSH_PATH, force_ipvx, options["-l"], options["-a"], options["-k"], options["-u"])
 			if options.has_key("ssh_options"):
@@ -764,14 +858,14 @@ def fence_login(options):
 				"are not in the spec file and must be installed separately." + "\n")
 				sys.exit(EC_GENERIC_ERROR)
 
-			result = conn.log_expect(options, [ options["-c"], "Are you sure you want to continue connecting (yes/no)?", "Enter passphrase for key '"+options["-k"]+"':" ], LOGIN_TIMEOUT)
+			result = conn.log_expect(options, [ options["-c"], "Are you sure you want to continue connecting (yes/no)?", "Enter passphrase for key '"+options["-k"]+"':" ], int(options["-y"]))
 			if result == 1:
 				conn.sendline("yes")
-				conn.log_expect(options, [ options["-c"], "Enter passphrase for key '"+options["-k"]+"':"] , LOGIN_TIMEOUT)
+				conn.log_expect(options, [ options["-c"], "Enter passphrase for key '"+options["-k"]+"':"] , int(options["-y"]))
 			if result != 0:
 				if options.has_key("-p"):
 					conn.sendline(options["-p"])
-					conn.log_expect(options, options["-c"], LOGIN_TIMEOUT)
+					conn.log_expect(options, options["-c"], int(options["-y"]))
 				else:
 					fail_usage("Failed: You have to enter passphrase (-p) for identity file")
 		else:
@@ -785,11 +879,11 @@ def fence_login(options):
 				"are not in the spec file and must be installed separately." + "\n")
 				sys.exit(EC_GENERIC_ERROR)
 
-			conn.log_expect(options, re_login, LOGIN_TIMEOUT)
+			conn.log_expect(options, re_login, int(options["-y"]))
 			conn.send(options["-l"] + login_eol)
-			conn.log_expect(options, re_pass, SHELL_TIMEOUT)
+			conn.log_expect(options, re_pass, int(options["-Y"]))
 			conn.send(options["-p"] + login_eol)
-			conn.log_expect(options, options["-c"], SHELL_TIMEOUT)
+			conn.log_expect(options, options["-c"], int(options["-Y"]))
 	except pexpect.EOF:
 		fail(EC_LOGIN_DENIED) 
 	except pexpect.TIMEOUT:
