@@ -352,7 +352,7 @@ do_read_direct(struct file *file, char *buf, size_t size, loff_t *offset,
 		goto out_gunlock;
 
 	if (gfs_is_stuffed(ip)) {
-		size_t mask = bdev_hardsect_size(inode->i_sb->s_bdev) - 1;
+		size_t mask = bdev_logical_block_size(inode->i_sb->s_bdev) - 1;
 
 		if (((*offset) & mask) || (((unsigned long)buf) & mask))
 			goto out_gunlock;
@@ -808,6 +808,7 @@ do_do_write_buf(struct file *file, char *buf, size_t size, loff_t *offset,
 	int alloc_required, journaled;
 	ssize_t count;
 	int error;
+	unsigned int posix = sdp->sd_args.ar_posix_acls ? 4 : 0;
 
 	journaled = gfs_is_jdata(ip);
 
@@ -845,7 +846,7 @@ do_do_write_buf(struct file *file, char *buf, size_t size, loff_t *offset,
 
 		error = gfs_trans_begin(sdp,
 					1 + al->al_rgd->rd_ri.ri_length +
-					ind_blocks +
+					ind_blocks + posix +
 					((journaled) ? data_blocks : 0), 1);
 		if (error)
 			goto fail_ipres;
@@ -853,7 +854,7 @@ do_do_write_buf(struct file *file, char *buf, size_t size, loff_t *offset,
 		/* Trans may require:
 		   A modified dinode. */
 
-		error = gfs_trans_begin(sdp,
+		error = gfs_trans_begin(sdp, posix +
 					1 + ((journaled) ? data_blocks : 0), 0);
 		if (error)
 			goto fail_ipres;
@@ -1731,7 +1732,8 @@ do_flock(struct file *file, int cmd, struct file_lock *fl)
 	int error = 0;
 
 	state = (fl->fl_type == F_WRLCK) ? LM_ST_EXCLUSIVE : LM_ST_SHARED;
-	flags = ((IS_SETLKW(cmd)) ? 0 : LM_FLAG_TRY) | GL_EXACT | GL_NOCACHE;
+	flags = ((IS_SETLKW(cmd)) ? 0 : LM_FLAG_TRY) | GL_EXACT | GL_NOCACHE
+		| GL_FLOCK;
 
 	down(&fp->f_fl_lock);
 
@@ -1739,21 +1741,19 @@ do_flock(struct file *file, int cmd, struct file_lock *fl)
 	if (gl) {
 		if (fl_gh->gh_state == state)
 			goto out;
-		gfs_glock_hold(gl);
 		flock_lock_file_wait(file,
-				     &(struct file_lock){.fl_type = F_UNLCK});		
-		gfs_glock_dq_uninit(fl_gh);
+				     &(struct file_lock){.fl_type = F_UNLCK});
+		gfs_glock_dq_wait(fl_gh);
+		gfs_holder_reinit(state, flags, fl_gh);
 	} else {
 		error = gfs_glock_get(ip->i_sbd,
 				      ip->i_num.no_formal_ino, &gfs_flock_glops,
 				      CREATE, &gl);
 		if (error)
 			goto out;
+		gfs_holder_init(gl, state, flags, fl_gh);
+		gfs_glock_put(gl);
 	}
-
-	gfs_holder_init(gl, state, flags, fl_gh);
-	gfs_glock_put(gl);
-
 	error = gfs_glock_nq(fl_gh);
 	if (error) {
 		gfs_holder_uninit(fl_gh);
