@@ -33,8 +33,8 @@ struct saved_metablock {
 };
 
 struct saved_metablock *savedata;
+struct gfs2_buffer_head *savebh;
 uint64_t last_fs_block, last_reported_block, blks_saved, total_out, pct;
-struct gfs2_block_list *blocklist = NULL;
 uint64_t journal_blocks[MAX_JOURNALS_SAVED];
 uint64_t gfs1_journal_size = 0; /* in blocks */
 int journals_found = 0;
@@ -50,15 +50,15 @@ extern void read_superblock(void);
  * returns: 0 if successful
  *          -1 if this isn't gfs metadata.
  */
-static int get_gfs_struct_info(char *gbuf, int *block_type, int *gstruct_len)
+static int get_gfs_struct_info(struct gfs2_buffer_head *lbh, int *block_type,
+			       int *gstruct_len)
 {
-	struct gfs2_meta_header mh, mhbuf;
+	struct gfs2_meta_header mh;
 
 	*block_type = 0;
 	*gstruct_len = sbd.bsize;
 
-	memcpy(&mhbuf, gbuf, sizeof(mhbuf));
-	gfs2_meta_header_in(&mh, (void *)&mhbuf);
+	gfs2_meta_header_in(&mh, lbh);
 	if (mh.mh_magic != GFS2_MAGIC)
 		return -1;
 
@@ -84,13 +84,7 @@ static int get_gfs_struct_info(char *gbuf, int *block_type, int *gstruct_len)
 		*gstruct_len = sbd.bsize; /*sizeof(struct gfs_leaf);*/
 		break;
 	case GFS2_METATYPE_JD:   /* 7 (journal data) */
-		/* GFS1 keeps indirect pointers in GFS2_METATYPE_JD blocks
-		   so we need to save the whole block.  For GFS2, we don't
-		   want to, or we might capture user data, which is bad.  */
-		if (gfs1)
-			*gstruct_len = sbd.bsize;
-		else
-			*gstruct_len = sizeof(struct gfs2_meta_header);
+		*gstruct_len = sbd.bsize;
 		break;
 	case GFS2_METATYPE_LH:   /* 8 (log header) */
 		if (gfs1)
@@ -139,7 +133,7 @@ static void warm_fuzzy_stuff(uint64_t wfsblock, int force, int save)
 		if (last_fs_block) {
 			printf("\r");
 			if (save) {
-				percent = (block * 100) / last_fs_block;
+				percent = (wfsblock * 100) / last_fs_block;
 				printf("%" PRIu64 " metadata blocks (%"
 				       PRIu64 "%%) processed, ", wfsblock,
 				       percent);
@@ -193,28 +187,15 @@ static int save_block(int fd, int out_fd, uint64_t blk)
 		return 0;
 	}
 	memset(savedata, 0, sizeof(struct saved_metablock));
-	if (lseek(fd, blk * sbd.bsize, SEEK_SET) != blk * sbd.bsize) {
-		fprintf(stderr, "bad seek: %s from %s:%d: "
-			"block %lld (0x%llx)\n", strerror(errno),
-			__FUNCTION__, __LINE__,
-			(unsigned long long)blk, (unsigned long long)blk);
-		exit(-1);
-	}
-	/* read in the block */
-	if (read(fd, savedata->buf, sbd.bsize) != sbd.bsize) {
-		fprintf(stderr, "bad read: %s from %s:%d: "
-			"block %lld (0x%llx)\n", strerror(errno),
-			__FUNCTION__, __LINE__,
-			(unsigned long long)blk, (unsigned long long)blk);
-		exit(-1);
-	}
+	savebh = bread(&sbd, blk);
+	memcpy(&savedata->buf, savebh->b_data, sbd.bsize);
 
 	/* If this isn't metadata and isn't a system file, we don't want it.
 	   Note that we're checking "block" here rather than blk.  That's
 	   because we want to know if the source inode's "block" is a system
 	   inode, not the block within the inode "blk". They may or may not
 	   be the same thing. */
-	if (get_gfs_struct_info(savedata->buf, &blktype, &blklen) &&
+	if (get_gfs_struct_info(savebh, &blktype, &blklen) &&
 	    !block_is_systemfile())
 		return 0; /* Not metadata, and not system file, so skip it */
 	trailing0 = 0;
@@ -311,15 +292,14 @@ static void save_indirect_blocks(int out_fd, osi_list_t *cur_list,
 		old_block = indir_block;
 		blktype = save_block(sbd.device_fd, out_fd, indir_block);
 		if (blktype == GFS2_METATYPE_EA) {
-			nbh = bread(&sbd.buf_list, indir_block);
+			nbh = bread(&sbd, indir_block);
 			save_ea_block(out_fd, nbh);
-			brelse(nbh, not_updated);
+			brelse(nbh);
 		}
 		if (height != hgt) { /* If not at max height */
-			nbh = bread(&sbd.buf_list, indir_block);
-			osi_list_add_prev(&nbh->b_altlist,
-					  cur_list);
-			brelse(nbh, not_updated);
+			nbh = bread(&sbd, indir_block);
+			osi_list_add_prev(&nbh->b_altlist, cur_list);
+			brelse(nbh);
 		}
 	} /* for all data on the indirect block */
 }
@@ -350,7 +330,7 @@ static void save_inode_data(int out_fd)
 
 	for (i = 0; i < GFS2_MAX_META_HEIGHT; i++)
 		osi_list_init(&metalist[i]);
-	metabh = bread(&sbd.buf_list, block);
+	metabh = bread(&sbd, block);
 	if (gfs1)
 		inode = inode_get(&sbd, metabh);
 	else
@@ -358,8 +338,15 @@ static void save_inode_data(int out_fd)
 	height = inode->i_di.di_height;
 	/* If this is a user inode, we don't follow to the file height.
 	   We stop one level less.  That way we save off the indirect
-	   pointer blocks but not the actual file contents. */
-	if (height && !block_is_systemfile())
+	   pointer blocks but not the actual file contents. The exception
+	   is directories, where the height represents the level at which
+	   the hash table exists, and we have to save the directory data. */
+	if (inode->i_di.di_flags & GFS2_DIF_EXHASH &&
+	    (S_ISDIR(inode->i_di.di_mode) ||
+	     (gfs1 && inode->i_di.__pad1 == GFS_FILE_DIR)))
+		height++;
+	else if (height && !block_is_systemfile() &&
+		 !S_ISDIR(inode->i_di.di_mode))
 		height--;
 	osi_list_add(&metabh->b_altlist, &metalist[0]);
         for (i = 1; i <= height; i++){
@@ -392,16 +379,17 @@ static void save_inode_data(int out_fd)
 	}
 	if (inode->i_di.di_eattr) { /* if this inode has extended attributes */
 		struct gfs2_meta_header mh;
+		struct gfs2_buffer_head *lbh;
 
-		metabh = bread(&sbd.buf_list, inode->i_di.di_eattr);
+		lbh = bread(&sbd, inode->i_di.di_eattr);
 		save_block(sbd.device_fd, out_fd, inode->i_di.di_eattr);
-		gfs2_meta_header_in(&mh, metabh->b_data);
+		gfs2_meta_header_in(&mh, lbh);
 		if (mh.mh_magic == GFS2_MAGIC &&
 		    mh.mh_type == GFS2_METATYPE_EA)
-			save_ea_block(out_fd, metabh);
+			save_ea_block(out_fd, lbh);
 		else if (mh.mh_magic == GFS2_MAGIC &&
 			 mh.mh_type == GFS2_METATYPE_IN)
-			save_indirect_blocks(out_fd, cur_list, metabh, 2, 2);
+			save_indirect_blocks(out_fd, cur_list, lbh, 2, 2);
 		else {
 			if (mh.mh_magic == GFS2_MAGIC) /* if it's metadata */
 				save_block(sbd.device_fd, out_fd,
@@ -415,15 +403,15 @@ static void save_inode_data(int out_fd)
 				(unsigned long long)block,
 				(unsigned long long)block);
 		}
-		brelse(metabh, not_updated);
+		brelse(lbh);
 	}
-	inode_put(inode, not_updated);
+	inode_put(&inode);
+	brelse(metabh);
 }
 
 static void get_journal_inode_blocks(void)
 {
 	int journal;
-	struct gfs2_buffer_head *bh;
 
 	journals_found = 0;
 	memset(journal_blocks, 0, sizeof(journal_blocks));
@@ -439,19 +427,18 @@ static void get_journal_inode_blocks(void)
 	for (journal = 0; ; journal++) { /* while journals exist */
 		uint64_t jblock;
 		int amt;
-		struct gfs2_dinode jdi;
 		struct gfs2_inode *j_inode = NULL;
 
 		if (gfs1) {
 			struct gfs_jindex ji;
 			char jbuf[sizeof(struct gfs_jindex)];
 
-			bh = bread(&sbd.buf_list, sbd1->sb_jindex_di.no_addr);
-			j_inode = gfs_inode_get(&sbd, bh);
+			j_inode = gfs_inode_read(&sbd,
+						 sbd1->sb_jindex_di.no_addr);
 			amt = gfs2_readi(j_inode, (void *)&jbuf,
 					 journal * sizeof(struct gfs_jindex),
 					 sizeof(struct gfs_jindex));
-			brelse(bh, not_updated);
+			inode_put(&j_inode);
 			if (!amt)
 				break;
 			gfs_jindex_in(&ji, jbuf);
@@ -461,21 +448,19 @@ static void get_journal_inode_blocks(void)
 			if (journal > indirect->ii[0].dirents - 3)
 				break;
 			jblock = indirect->ii[0].dirent[journal + 2].block;
-			bh = bread(&sbd.buf_list, jblock);
-			j_inode = inode_get(&sbd, bh);
-			gfs2_dinode_in(&jdi, bh->b_data);
-			inode_put(j_inode, not_updated);
 		}
 		journal_blocks[journals_found++] = jblock;
 	}
 }
 
-static int next_rg_freemeta(struct rgrp_list *rgd, uint64_t *nrfblock, int first)
+static int next_rg_freemeta(struct gfs2_sbd *sdp, struct rgrp_list *rgd,
+			    uint64_t *nrfblock, int first)
 {
 	struct gfs2_bitmap *bits = NULL;
 	uint32_t length = rgd->ri.ri_length;
 	uint32_t blk = (first)? 0: (uint32_t)((*nrfblock+1)-rgd->ri.ri_data0);
 	int i;
+	struct gfs2_buffer_head *lbh;
 
 	if(!first && (*nrfblock < rgd->ri.ri_data0)) {
 		log_err("next_rg_freemeta:  Start block is outside rgrp "
@@ -490,9 +475,11 @@ static int next_rg_freemeta(struct rgrp_list *rgd, uint64_t *nrfblock, int first
 	}
 	for(; i < length; i++){
 		bits = &rgd->bits[i];
-		blk = gfs2_bitfit((unsigned char *)rgd->bh[i]->b_data +
+		lbh = bread(sdp, rgd->ri.ri_addr + i);
+		blk = gfs2_bitfit((unsigned char *)lbh->b_data +
 				  bits->bi_offset, bits->bi_len, blk,
 				  GFS2_BLKST_UNLINKED);
+		brelse(lbh);
 		if(blk != BFITNOENT){
 			*nrfblock = blk + (bits->bi_start * GFS2_NBBY) +
 				rgd->ri.ri_data0;
@@ -510,10 +497,9 @@ void savemeta(char *out_fn, int saveoption)
 	int out_fd;
 	int slow;
 	osi_list_t *tmp;
-	uint64_t memreq;
 	int rgcount;
 	uint64_t jindex_block;
-	struct gfs2_buffer_head *bh;
+	struct gfs2_buffer_head *lbh;
 
 	slow = (saveoption == 1);
 	sbd.md.journals = 1;
@@ -550,8 +536,6 @@ void savemeta(char *out_fn, int saveoption)
 			exit(-1);
 		}
 		osi_list_init(&sbd.rglist);
-		init_buf_list(&sbd, &sbd.buf_list, 128 << 20);
-		init_buf_list(&sbd, &sbd.nvbuf_list, 0xffffffff);
 		if (!gfs1)
 			sbd.sd_sb.sb_bsize = GFS2_DEFAULT_BSIZE;
 		if (compute_constants(&sbd)) {
@@ -584,41 +568,36 @@ void savemeta(char *out_fn, int saveoption)
 	       last_fs_block, sbd.bsize);
 	if (!slow) {
 		if (gfs1) {
-			sbd.md.riinode =
-				gfs2_load_inode(&sbd,
+			sbd.md.riinode = inode_read(&sbd,
 						sbd1->sb_rindex_di.no_addr);
 			jindex_block = sbd1->sb_jindex_di.no_addr;
 		} else {
 			sbd.master_dir =
-				gfs2_load_inode(&sbd,
-						sbd.sd_sb.sb_master_dir.no_addr);
+				inode_read(&sbd,
+					sbd.sd_sb.sb_master_dir.no_addr);
 
 			slow = gfs2_lookupi(sbd.master_dir, "rindex", 6, 
 					    &sbd.md.riinode);
 			jindex_block = masterblock("jindex");
 		}
-		bh = bread(&sbd.buf_list, jindex_block);
-		gfs2_dinode_in(&di, bh->b_data);
+		lbh = bread(&sbd, jindex_block);
+		gfs2_dinode_in(&di, lbh);
 		if (!gfs1)
-			do_dinode_extended(&di, bh->b_data);
-		brelse(bh, not_updated);
+			do_dinode_extended(&di, lbh);
+		brelse(lbh);
 	}
 	if (!slow) {
+		int sane;
+
 		printf("Reading resource groups...");
 		fflush(stdout);
 		if (gfs1)
-			slow = gfs1_ri_update(&sbd, 0, &rgcount);
+			slow = gfs1_ri_update(&sbd, 0, &rgcount, 0);
 		else
-			slow = ri_update(&sbd, 0, &rgcount);
+			slow = ri_update(&sbd, 0, &rgcount, &sane);
 		printf("Done.\n\n");
 		fflush(stdout);
 	}
-	if (!slow) {
-		blocklist = gfs2_block_list_create(&sbd, last_fs_block + 1,
-						   &memreq);
-		if (!blocklist)
-			slow = TRUE;
-	}
 	get_journal_inode_blocks();
 	if (!slow) {
 		/* Save off the superblock */
@@ -645,7 +624,7 @@ void savemeta(char *out_fn, int saveoption)
 		for (tmp = sbd.rglist.next; tmp != &sbd.rglist;
 		     tmp = tmp->next){
 			struct rgrp_list *rgd;
-			int i, first;
+			int first;
 
 			rgd = osi_list_entry(tmp, struct rgrp_list, list);
 			slow = gfs2_rgrp_read(&sbd, rgd);
@@ -655,12 +634,6 @@ void savemeta(char *out_fn, int saveoption)
 				  (unsigned long long)rgd->ri.ri_addr,
 				  (unsigned long long)rgd->ri.ri_addr,
 				  rgd->ri.ri_length);
-			for (i = 0; i < rgd->ri.ri_length; i++) {
-				if(gfs2_block_set(&sbd, blocklist,
-						  rgd->ri.ri_addr + i,
-						  gfs2_meta_other))
-					break;
-			}
 			first = 1;
 			/* Save off the rg and bitmaps */
 			for (block = rgd->ri.ri_addr;
@@ -672,7 +645,7 @@ void savemeta(char *out_fn, int saveoption)
 			if (saveoption != 2) {
 				int blktype;
 
-				while (!gfs2_next_rg_meta(rgd, &block, first)) {
+				while (!gfs2_next_rg_meta(rgd, &block, first)){
 					warm_fuzzy_stuff(block, FALSE, TRUE);
 					blktype = save_block(sbd.device_fd,
 							     out_fd, block);
@@ -683,13 +656,14 @@ void savemeta(char *out_fn, int saveoption)
 				/* Save off the free/unlinked meta blocks too.
 				   If we don't, we may run into metadata
 				   allocation issues. */
-				while (!next_rg_freemeta(rgd, &block, first)) {
+				while (!next_rg_freemeta(&sbd, rgd, &block,
+							 first)) {
 					blktype = save_block(sbd.device_fd,
 							     out_fd, block);
 					first = 0;
 				}
 			}
-			gfs2_rgrp_relse(rgd, not_updated);
+			gfs2_rgrp_relse(rgd);
 		}
 	}
 	if (slow) {
@@ -698,8 +672,6 @@ void savemeta(char *out_fn, int saveoption)
 		}
 	}
 	/* Clean up */
-	if (blocklist)
-		gfs2_block_list_destroy(&sbd, blocklist);
 	/* There may be a gap between end of file system and end of device */
 	/* so we tell the user that we've processed everything. */
 	block = last_fs_block;
@@ -711,10 +683,27 @@ void savemeta(char *out_fn, int saveoption)
 	exit(0);
 }
 
+/**
+ * anthropomorphize - make a uint64_t number more human
+ */
+static const char *anthropomorphize(unsigned long long inhuman_value)
+{
+	const char *symbols = " KMGTPE";
+	int i;
+	unsigned long long val = inhuman_value;
+	static char out_val[32];
+
+	memset(out_val, 0, sizeof(out_val));
+	for (i = 0; i < 6 && val > 1024; i++)
+		val /= 1024;
+	sprintf(out_val, "%llu%c", val, symbols[i]);
+	return out_val;
+}
+
 static int restore_data(int fd, int in_fd, int printblocksonly)
 {
 	size_t rs;
-	uint64_t buf64, writes = 0;
+	uint64_t buf64, writes = 0, highest_valid_block = 0;
 	uint16_t buf16;
 	int first = 1, pos;
 	char rdbuf[256];
@@ -748,6 +737,8 @@ static int restore_data(int fd, int in_fd, int printblocksonly)
 	blks_saved = total_out = 0;
 	last_fs_block = 0;
 	while (TRUE) {
+		struct gfs2_buffer_head dummy_bh;
+
 		memset(savedata, 0, sizeof(struct saved_metablock));
 		rs = read(in_fd, &buf64, sizeof(uint64_t));
 		if (!rs)
@@ -770,93 +761,107 @@ static int restore_data(int fd, int in_fd, int printblocksonly)
 		}
 		rs = read(in_fd, &buf16, sizeof(uint16_t));
 		savedata->siglen = be16_to_cpu(buf16);
-		if (savedata->siglen <= sizeof(savedata->buf)) {
-			if (savedata->siglen &&
-			    read(in_fd, savedata->buf, savedata->siglen) !=
-			    savedata->siglen) {
-				fprintf(stderr, "read error: %s from %s:%d: "
-					"block %lld (0x%llx)\n",
+		if (savedata->siglen > sizeof(savedata->buf)) {
+			fprintf(stderr, "\nBad record length: %d for block #%"
+				PRIu64 " (0x%" PRIx64").\n", savedata->siglen,
+				savedata->blk, savedata->blk);
+			return -1;
+		}
+		if (savedata->siglen &&
+		    read(in_fd, savedata->buf, savedata->siglen) !=
+		    savedata->siglen) {
+			fprintf(stderr, "read error: %s from %s:%d: "
+				"block %lld (0x%llx)\n",
+				strerror(errno), __FUNCTION__, __LINE__,
+				(unsigned long long)savedata->blk,
+				(unsigned long long)savedata->blk);
+			exit(-1);
+		}
+		if (first) {
+			struct gfs2_sb bufsb;
+
+			dummy_bh.b_data = (char *)&bufsb;
+			memcpy(&bufsb, savedata->buf, sizeof(bufsb));
+			gfs2_sb_in(&sbd.sd_sb, &dummy_bh);
+			sbd1 = (struct gfs_sb *)&sbd.sd_sb;
+			if (sbd1->sb_fs_format == GFS_FORMAT_FS &&
+			    sbd1->sb_header.mh_type ==
+			    GFS_METATYPE_SB &&
+			    sbd1->sb_header.mh_format ==
+			    GFS_FORMAT_SB &&
+			    sbd1->sb_multihost_format ==
+			    GFS_FORMAT_MULTI) {
+				gfs1 = TRUE;
+			} else if (check_sb(&sbd.sd_sb)) {
+				fprintf(stderr,"Error: Invalid superblock data.\n");
+				return -1;
+			}
+			sbd.bsize = sbd.sd_sb.sb_bsize;
+			if (!printblocksonly) {
+				last_fs_block =
+					lseek(fd, 0, SEEK_END) / sbd.bsize;
+				printf("There are %" PRIu64 " blocks of " \
+				       "%u bytes in the destination"	\
+				       " file system.\n\n",
+				       last_fs_block, sbd.bsize);
+			} else {
+				printf("This is %s metadata\n", gfs1 ?
+				       "gfs (not gfs2)" : "gfs2");
+			}
+			first = 0;
+		}
+		bh = &dummy_bh;
+		bh->b_data = savedata->buf;
+		if (printblocksonly) {
+			block = savedata->blk;
+			if (block > highest_valid_block)
+				highest_valid_block = block;
+			if (printblocksonly > 1 && printblocksonly == block) {
+				block_in_mem = block;
+				display(0);
+				return 0;
+			} else if (printblocksonly == 1) {
+				print_gfs2("%d (l=0x%x): ", blks_saved,
+					   savedata->siglen);
+				display_block_type(TRUE);
+			}
+		} else {
+			warm_fuzzy_stuff(savedata->blk, FALSE, FALSE);
+			if (savedata->blk >= last_fs_block) {
+				printf("\nOut of space on the destination "
+				       "device; quitting.\n");
+				break;
+			}
+			if (lseek(fd, savedata->blk * sbd.bsize, SEEK_SET) !=
+			    savedata->blk * sbd.bsize) {
+				fprintf(stderr, "bad seek: %s from %s:"
+					"%d: block %lld (0x%llx)\n",
 					strerror(errno), __FUNCTION__,
 					__LINE__,
 					(unsigned long long)savedata->blk,
 					(unsigned long long)savedata->blk);
 				exit(-1);
 			}
-			if (first) {
-				struct gfs2_sb bufsb;
-
-				memcpy(&bufsb, savedata->buf, sizeof(bufsb));
-				gfs2_sb_in(&sbd.sd_sb, (void *)&bufsb);
-				sbd1 = (struct gfs_sb *)&sbd.sd_sb;
-				if (sbd1->sb_fs_format == GFS_FORMAT_FS &&
-				    sbd1->sb_header.mh_type ==
-				    GFS_METATYPE_SB &&
-				    sbd1->sb_header.mh_format ==
-				    GFS_FORMAT_SB &&
-				    sbd1->sb_multihost_format ==
-				    GFS_FORMAT_MULTI)
-					;
-				else if (check_sb(&sbd.sd_sb)) {
-					fprintf(stderr,"Error: Invalid superblock data.\n");
-					return -1;
-				}
-				sbd.bsize = sbd.sd_sb.sb_bsize;
-				if (!printblocksonly) {
-					last_fs_block =
-						lseek(fd, 0, SEEK_END) /
-						sbd.bsize;
-					printf("There are %" PRIu64 " blocks of " \
-					       "%u bytes in the destination" \
-					       " file system.\n\n",
-					       last_fs_block, sbd.bsize);
-				}
-				first = 0;
+			if (write(fd, savedata->buf, sbd.bsize) != sbd.bsize) {
+				fprintf(stderr, "write error: %s from "
+					"%s:%d: block %lld (0x%llx)\n",
+					strerror(errno), __FUNCTION__,
+					__LINE__,
+					(unsigned long long)savedata->blk,
+					(unsigned long long)savedata->blk);
+				exit(-1);
 			}
-			if (printblocksonly) {
-				print_gfs2("%d (l=0x%x): ", blks_saved,
-					   savedata->siglen);
-				block = savedata->blk;
-				display_block_type(savedata->buf, TRUE);
-			} else {
-				warm_fuzzy_stuff(savedata->blk, FALSE, FALSE);
-				if (savedata->blk >= last_fs_block) {
-					printf("\nOut of space on the destination "
-					       "device; quitting.\n");
-					break;
-				}
-				if (lseek(fd, savedata->blk * sbd.bsize,
-					  SEEK_SET) !=
-				    savedata->blk * sbd.bsize) {
-					fprintf(stderr, "bad seek: %s from %s:"
-						"%d: block %lld (0x%llx)\n",
-						strerror(errno), __FUNCTION__,
-						__LINE__, (unsigned long long)
-						savedata->blk,
-						(unsigned long long)
-						savedata->blk);
-					exit(-1);
-				}
-				if (write(fd, savedata->buf, sbd.bsize) !=
-				    sbd.bsize) {
-					fprintf(stderr, "write error: %s from "
-						"%s:%d: block %lld (0x%llx)\n",
-						strerror(errno),
-						__FUNCTION__, __LINE__,
-						(unsigned long long)savedata->blk,
-						(unsigned long long)savedata->blk);
-					exit(-1);
-				}
-				writes++;
-			}
-			blks_saved++;
-		} else {
-			fprintf(stderr, "\nBad record length: %d for block #%"
-				PRIu64".\n", savedata->siglen, savedata->blk);
-			return -1;
+			writes++;
 		}
+		blks_saved++;
 	}
 	if (!printblocksonly)
 		warm_fuzzy_stuff(savedata->blk, TRUE, FALSE);
+	else
+		printf("File system size: %lld (0x%llx) blocks, aka %sB\n",
+		       (unsigned long long)highest_valid_block,
+		       (unsigned long long)highest_valid_block,
+		       anthropomorphize(highest_valid_block * sbd.bsize));
 	return 0;
 }
 
@@ -868,7 +873,7 @@ static void complain(const char *complaint)
 }
 
 void restoremeta(const char *in_fn, const char *out_device,
-		 int printblocksonly)
+		 uint64_t printblocksonly)
 {
 	int in_fd, error;
 
@@ -887,7 +892,9 @@ void restoremeta(const char *in_fn, const char *out_device,
 		if (sbd.device_fd < 0)
 			die("Can't open destination file system %s: %s\n",
 			    out_device, strerror(errno));
-	}
+	} else if (out_device) /* for printsavedmeta, the out_device is an
+				  optional block no */
+		printblocksonly = check_keywords(out_device);
 	savedata = malloc(sizeof(struct saved_metablock));
 	if (!savedata)
 		die("Can't allocate memory for the restore operation.\n");
