@@ -15,19 +15,24 @@
 #include "libgfs2.h"
 #include "fsck.h"
 #include "osi_list.h"
+#include "metawalk.h"
+#include "util.h"
 
 struct gfs2_options opts = {0};
-struct gfs2_inode *lf_dip; /* Lost and found directory inode */
-osi_list_t dir_hash[FSCK_HASH_SIZE];
-osi_list_t inode_hash[FSCK_HASH_SIZE];
-struct gfs2_block_list *bl = NULL;
+struct gfs2_inode *lf_dip = NULL; /* Lost and found directory inode */
+struct gfs2_bmap *bl = NULL;
 uint64_t last_fs_block, last_reported_block = -1;
+int64_t last_reported_fblock = -1000000;
 int skip_this_pass = FALSE, fsck_abort = FALSE;
 int errors_found = 0, errors_corrected = 0;
 const char *pass = "";
 uint64_t last_data_block;
 uint64_t first_data_block;
 int preen = 0, force_check = 0;
+struct osi_root dup_blocks = (struct osi_root) { NULL, };
+struct osi_root dirtree = (struct osi_root) { NULL, };
+struct osi_root inodetree = (struct osi_root) { NULL, };
+int dups_found = 0, dups_found_first = 0;
 
 /* This function is for libgfs2's sake.                                      */
 void print_it(const char *label, const char *fmt, const char *fmt2, ...)
@@ -139,124 +144,73 @@ static void interrupt(int sig)
 	}
 }
 
-/* Check system inode and verify it's marked "in use" in the bitmap:       */
-/* Should work for all system inodes: root, master, jindex, per_node, etc. */
-static int check_system_inode(struct gfs2_inode *sysinode, const char *filename,
-		       int builder(struct gfs2_sbd *sbp),
-		       enum gfs2_mark_block mark)
+static void check_statfs(struct gfs2_sbd *sdp)
 {
-	uint64_t iblock = 0;
-	struct dir_status ds = {0};
+	osi_list_t *tmp;
+	struct rgrp_list *rgd;
+	struct gfs2_rindex *ri;
+	struct gfs2_statfs_change sc;
+	char buf[sizeof(struct gfs2_statfs_change)];
+	int count;
 
-	log_info( _("Checking system inode '%s'\n"), filename);
-	if (sysinode) {
-		/* Read in the system inode, look at its dentries, and start
-		 * reading through them */
-		iblock = sysinode->i_di.di_num.no_addr;
-		log_info( _("System inode for '%s' is located at block %"
-			 PRIu64 " (0x%" PRIx64 ")\n"), filename,
-			 iblock, iblock);
-		
-		/* FIXME: check this block's validity */
+	/* Read the current statfs values */
+	count = gfs2_readi(sdp->md.statfs, buf, 0,
+			   sdp->md.statfs->i_di.di_size);
+	if (count == sizeof(struct gfs2_statfs_change))
+		gfs2_statfs_change_in(&sc, buf);
 
-		if(gfs2_block_check(sysinode->i_sbd, bl, iblock, &ds.q)) {
-			log_crit( _("Can't get %s inode block %" PRIu64 " (0x%"
-				 PRIx64 ") from block list\n"), filename,
-				 iblock, iblock);
-			return -1;
-		}
-		/* If the inode exists but the block is marked      */
-		/* free, we might be recovering from a corrupt      */
-		/* bitmap.  In that case, don't rebuild the inode.  */
-		/* Just reuse the inode and fix the bitmap.         */
-		if (ds.q.block_type == gfs2_block_free) {
-			log_info( _("The inode exists but the block is not marked 'in use'; fixing it.\n"));
-			gfs2_block_set(sysinode->i_sbd, bl,
-				       sysinode->i_di.di_num.no_addr,
-				       mark);
-			ds.q.block_type = mark;
-			if (mark == gfs2_inode_dir)
-				add_to_dir_list(sysinode->i_sbd,
-						sysinode->i_di.di_num.no_addr);
-		}
-	}
-	else
-		log_info( _("System inode for '%s' is missing.\n"), filename);
-	/* If there are errors with the inode here, we need to
-	 * create a new inode and get it all setup - of course,
-	 * everything will be in lost+found then, but we *need* our
-	 * system inodes before we can do any of that. */
-	if(!sysinode || ds.q.block_type != mark) {
-		log_err( _("Invalid or missing %s system inode.\n"), filename);
-		errors_found++;
-		if (query(&opts, _("Create new %s system inode? (y/n) "),
-			  filename)) {
-			errors_corrected++;
-			builder(sysinode->i_sbd);
-			gfs2_block_set(sysinode->i_sbd, bl,
-				       sysinode->i_di.di_num.no_addr,
-				       mark);
-			ds.q.block_type = mark;
-			if (mark == gfs2_inode_dir)
-				add_to_dir_list(sysinode->i_sbd,
-						sysinode->i_di.di_num.no_addr);
-		}
-		else {
-			log_err( _("Cannot continue without valid %s inode\n"),
-				filename);
-			return -1;
-		}
+	/* Calculate the real values from the rgrp information */
+	sdp->blks_total = 0;
+	sdp->blks_alloced = 0;
+	sdp->dinodes_alloced = 0;
+
+	for (tmp = sdp->rglist.next; tmp != &sdp->rglist; tmp = tmp->next) {
+		rgd = osi_list_entry(tmp, struct rgrp_list, list);
+		ri = &rgd->ri;
+		sdp->blks_total += ri->ri_data;
+		sdp->blks_alloced += (ri->ri_data - rgd->rg.rg_free);
+		sdp->dinodes_alloced += rgd->rg.rg_dinodes;
 	}
 
-	return 0;
-}
+	/* See if they match */
+	if (sc.sc_total == sdp->blks_total &&
+	    sc.sc_free == (sdp->blks_total - sdp->blks_alloced) &&
+	    sc.sc_dinodes == sdp->dinodes_alloced) {
+		log_info( _("The statfs file is accurate.\n"));
+		return;
+	}
+	log_err( _("The statfs file is wrong:\n\n"));
+	log_err( _("Current statfs values:\n"));
+	log_err( _("blocks:  %lld (0x%llx)\n"),
+		 (unsigned long long)sc.sc_total,
+		 (unsigned long long)sc.sc_total);
+	log_err( _("free:    %lld (0x%llx)\n"),
+		 (unsigned long long)sc.sc_free,
+		 (unsigned long long)sc.sc_free);
+	log_err( _("dinodes: %lld (0x%llx)\n\n"),
+		 (unsigned long long)sc.sc_dinodes,
+		 (unsigned long long)sc.sc_dinodes);
 
-static int check_system_inodes(struct gfs2_sbd *sdp)
-{
-	/*******************************************************************
-	 *******  Check the system inode integrity             *************
-	 *******************************************************************/
-	if (check_system_inode(sdp->master_dir, "master", build_master,
-			       gfs2_inode_dir)) {
-		stack;
-		return -1;
+	log_err( _("Calculated statfs values:\n"));
+	log_err( _("blocks:  %lld (0x%llx)\n"),
+		 (unsigned long long)sdp->blks_total,
+		 (unsigned long long)sdp->blks_total);
+	log_err( _("free:    %lld (0x%llx)\n"),
+		 (unsigned long long)(sdp->blks_total - sdp->blks_alloced),
+		 (unsigned long long)(sdp->blks_total - sdp->blks_alloced));
+	log_err( _("dinodes: %lld (0x%llx)\n"),
+		 (unsigned long long)sdp->dinodes_alloced,
+		 (unsigned long long)sdp->dinodes_alloced);
+
+	errors_found++;
+	if (!query( _("Okay to fix the master statfs file? (y/n)"))) {
+		log_err( _("The statfs file was not fixed.\n"));
+		return;
 	}
-	if (check_system_inode(sdp->md.rooti, "root", build_root,
-			       gfs2_inode_dir)) {
-		stack;
-		return -1;
-	}
-	if (check_system_inode(sdp->md.inum, "inum", build_inum,
-			       gfs2_inode_file)) {
-		stack;
-		return -1;
-	}
-	if (check_system_inode(sdp->md.statfs, "statfs", build_statfs,
-			       gfs2_inode_file)) {
-		stack;
-		return -1;
-	}
-	if (check_system_inode(sdp->md.jiinode, "jindex", build_jindex,
-			       gfs2_inode_dir)) {
-		stack;
-		return -1;
-	}
-	if (check_system_inode(sdp->md.riinode, "rindex", build_rindex,
-			       gfs2_inode_file)) {
-		stack;
-		return -1;
-	}
-	if (check_system_inode(sdp->md.qinode, "quota", build_quota,
-			       gfs2_inode_file)) {
-		stack;
-		return -1;
-	}
-	if (check_system_inode(sdp->md.pinode, "per_node", build_per_node,
-			       gfs2_inode_dir)) {
-		stack;
-		return -1;
-	}
-	return 0;
+
+	do_init_statfs(sdp);
+	log_err( _("The statfs file was fixed.\n"));
+	errors_corrected++;
 }
 
 int main(int argc, char **argv)
@@ -264,7 +218,6 @@ int main(int argc, char **argv)
 	struct gfs2_sbd sb;
 	struct gfs2_sbd *sbp = &sb;
 	int j;
-	enum update_flags update_sys_files;
 	int error = 0;
 	int all_clean = 0;
 
@@ -299,9 +252,6 @@ int main(int argc, char **argv)
 	else
 		log_notice( _("Pass1 complete      \n"));
 
-	/* Make sure the system inodes are okay & represented in the bitmap. */
-	check_system_inodes(sbp);
-
 	if (!fsck_abort) {
 		last_reported_block = 0;
 		pass = "pass 1b";
@@ -383,25 +333,26 @@ int main(int argc, char **argv)
 	} else {
 		error = FSCK_CANCELED;
 	}
-	update_sys_files = (opts.no ? not_updated : updated);
-	/* Free up our system inodes */
-	inode_put(sbp->md.inum, update_sys_files);
-	inode_put(sbp->md.statfs, update_sys_files);
-	for (j = 0; j < sbp->md.journals; j++)
-		inode_put(sbp->md.journal[j], update_sys_files);
-	inode_put(sbp->md.jiinode, update_sys_files);
-	inode_put(sbp->md.riinode, update_sys_files);
-	inode_put(sbp->md.qinode, update_sys_files);
-	inode_put(sbp->md.pinode, update_sys_files);
-	inode_put(sbp->md.rooti, update_sys_files);
-	inode_put(sbp->master_dir, update_sys_files);
-	if (lf_dip)
-		inode_put(lf_dip, update_sys_files);
 
-	if (!opts.no)
+	check_statfs(sbp);
+
+	/* Free up our system inodes */
+	inode_put(&sbp->md.inum);
+	inode_put(&sbp->md.statfs);
+	for (j = 0; j < sbp->md.journals; j++)
+		inode_put(&sbp->md.journal[j]);
+	inode_put(&sbp->md.jiinode);
+	inode_put(&sbp->md.riinode);
+	inode_put(&sbp->md.qinode);
+	inode_put(&sbp->md.pinode);
+	inode_put(&sbp->md.rooti);
+	inode_put(&sbp->master_dir);
+	if (lf_dip)
+		inode_put(&lf_dip);
+
+	if (!opts.no && errors_corrected)
 		log_notice( _("Writing changes to disk\n"));
-	bsync(&sbp->buf_list);
-	bsync(&sbp->nvbuf_list);
+	fsync(sbp->device_fd);
 	destroy(sbp);
 	log_notice( _("gfs2_fsck complete    \n"));
 
