@@ -443,6 +443,35 @@ static void query_dump_debug(int fd)
 	do_write(fd, dump_buf, len);
 }
 
+static void query_dump_log_plock(int fd)
+{
+	struct dlmc_header h;
+	int extra_len;
+	int len;
+
+	/* in the case of dump_wrap, extra_len will go in two writes,
+	   first the log tail, then the log head */
+	if (log_plock_wrap)
+		extra_len = DLMC_DUMP_SIZE;
+	else
+		extra_len = log_plock_point;
+
+	init_header(&h, DLMC_CMD_DUMP_LOG_PLOCK, NULL, 0, extra_len);
+	do_write(fd, &h, sizeof(h));
+
+	if (log_plock_wrap) {
+		len = DLMC_DUMP_SIZE - log_plock_point;
+		do_write(fd, log_plock_buf + log_plock_point, len);
+		len = log_plock_point;
+	} else
+		len = log_plock_point;
+
+	/* NUL terminate the debug string */
+	log_plock_buf[log_plock_point] = '\0';
+
+	do_write(fd, log_plock_buf, len);
+}
+
 static void query_dump_plocks(int fd, char *name)
 {
 	struct lockspace *ls;
@@ -804,6 +833,9 @@ static void *process_queries(void *arg)
 		case DLMC_CMD_DUMP_DEBUG:
 			query_dump_debug(f);
 			break;
+		case DLMC_CMD_DUMP_LOG_PLOCK:
+			query_dump_log_plock(f);
+			break;
 		case DLMC_CMD_DUMP_PLOCKS:
 			query_dump_plocks(f, h.name);
 			break;
@@ -902,7 +934,7 @@ static void loop(void)
 
 	setup_monitor();
 
-	rv = setup_configfs();
+	rv = setup_configfs();		/* calls update_cluster() */
 	if (rv < 0)
 		goto out;
 
@@ -938,10 +970,10 @@ static void loop(void)
 	log_debug("group_mode %d compat %d", group_mode, cfgd_groupd_compat);
 
 	if (group_mode == GROUP_LIBCPG) {
-		rv = setup_cpg();
+		rv = setup_cpg_daemon();
 		if (rv < 0)
 			goto out;
-		client_add(rv, process_cpg, cluster_dead);
+		client_add(rv, process_cpg_daemon, cluster_dead);
 
 		rv = set_protocol();
 		if (rv < 0)
@@ -1011,13 +1043,20 @@ static void loop(void)
 			}
 			poll_timeout = 1000;
 		}
+
+		if (poll_drop_plock) {
+			drop_resources_all();
+			if (poll_drop_plock)
+				poll_timeout = 1000;
+		}
+
 		query_unlock();
 	}
  out:
 	if (cfgd_groupd_compat)
 		close_groupd();
 	if (group_mode == GROUP_LIBCPG) {
-		close_cpg();
+		close_cpg_daemon();
 		close_plocks();
 	}
 	clear_configfs();
@@ -1082,6 +1121,9 @@ static void print_usage(void)
 	printf("  -D		Enable debugging to stderr and don't fork\n");
 	printf("  -L		Enable debugging to log file\n");
 	printf("  -K		Enable kernel dlm debugging messages\n");
+	printf("  -r <num>	dlm kernel lowcomms protocol, 0 tcp, 1 sctp, 2 detect\n");
+	printf("                2 selects tcp if corosync rrp_mode is \"none\", otherwise sctp\n");
+	printf("                Default is 2\n");
 	printf("  -g <num>	groupd compatibility mode, 0 off, 1 on, 2 detect\n");
 	printf("		0: use libcpg, no backward compat, best performance\n");
 	printf("		1: use libgroup for compat with cluster2/rhel5\n");
@@ -1111,17 +1153,13 @@ static void print_usage(void)
 	printf("  -V		Print program version information, then exit\n");
 }
 
-#define OPTION_STRING "LDKg:f:q:d:p:Pl:o:t:c:a:hV"
+#define OPTION_STRING "LDKg:f:q:d:p:Pl:o:t:c:a:hVr:"
 
 static void read_arguments(int argc, char **argv)
 {
 	int cont = 1;
 	int optchar;
 
-	/* we don't allow these to be set on command line, should we? */
-	optk_timewarn = 0;
-	optk_timewarn = 0;
-
 	while (cont) {
 		optchar = getopt(argc, argv, OPTION_STRING);
 
@@ -1146,6 +1184,11 @@ static void read_arguments(int argc, char **argv)
 			cfgk_debug = 1;
 			break;
 
+		case 'r':
+			optk_protocol = 1;
+			cfgk_protocol = atoi(optarg);
+			break;
+
 		case 'f':
 			optd_enable_fencing = 1;
 			cfgd_enable_fencing = atoi(optarg);
@@ -1303,6 +1346,22 @@ void daemon_dump_save(void)
 	}
 }
 
+void log_plock_save(void)
+{
+	int len, i;
+
+	len = strlen(log_plock_line);
+
+	for (i = 0; i < len; i++) {
+		log_plock_buf[log_plock_point++] = log_plock_line[i];
+
+		if (log_plock_point == DLMC_DUMP_SIZE) {
+			log_plock_point = 0;
+			log_plock_wrap = 1;
+		}
+	}
+}
+
 int daemon_debug_opt;
 int daemon_quit;
 int cluster_down;
@@ -1310,15 +1369,12 @@ int poll_fencing;
 int poll_quorum;
 int poll_fs;
 int poll_ignore_plock;
+int poll_drop_plock;
 int plock_fd;
 int plock_ci;
 struct list_head lockspaces;
 int cluster_quorate;
 int our_nodeid;
-char daemon_debug_buf[256];
-char dump_buf[DLMC_DUMP_SIZE];
-int dump_point;
-int dump_wrap;
 char plock_dump_buf[DLMC_DUMP_SIZE];
 int plock_dump_len;
 int group_mode;
@@ -1326,6 +1382,14 @@ uint32_t control_minor;
 uint32_t monitor_minor;
 uint32_t plock_minor;
 uint32_t old_plock_minor;
+char daemon_debug_buf[256];
+char dump_buf[DLMC_DUMP_SIZE];
+int dump_point;
+int dump_wrap;
+char log_plock_line[256];
+char log_plock_buf[DLMC_DUMP_SIZE];
+int log_plock_point;
+int log_plock_wrap;
 
 /* was a config value set on command line?, 0 or 1.
    optk is a kernel option, optd is a daemon option */
@@ -1351,7 +1415,7 @@ int optd_drop_resources_age;
 
 int cfgk_debug                  = -1;
 int cfgk_timewarn               = -1;
-int cfgk_protocol               = -1;
+int cfgk_protocol               = PROTO_DETECT;
 int cfgd_groupd_compat          = DEFAULT_GROUPD_COMPAT;
 int cfgd_debug_logfile		= DEFAULT_DEBUG_LOGFILE;
 int cfgd_enable_fencing         = DEFAULT_ENABLE_FENCING;
