@@ -888,6 +888,7 @@ void process_connection(int ci)
 	struct gfsc_mount_args empty;
 	struct gfsc_mount_args *ma;
 	char *extra = NULL;
+	char *fsname;
 	int rv, extra_len;
 
 	rv = do_read(client[ci].fd, &h, sizeof(h));
@@ -943,10 +944,16 @@ void process_connection(int ci)
 		break;
 
 	case GFSC_CMD_FS_LEAVE:
+		fsname = strstr(ma->table, ":") + 1;
+		if (!fsname) {
+			log_error("process_connection no fsname in table");
+			goto out;
+		}
+
 		if (group_mode == GROUP_LIBGROUP)
-			do_leave_old(ma->table, h.data);
+			do_leave_old(fsname, h.data);
 		else
-			do_leave(ma->table, h.data);
+			do_leave(fsname, h.data);
 		break;
 
 	case GFSC_CMD_FS_MOUNT_DONE:
@@ -1131,10 +1138,12 @@ static void loop(void)
 		goto out;
 	client_add(rv, process_listener, NULL);
 
-	rv = setup_cman();
+	rv = setup_cluster();
 	if (rv < 0)
 		goto out;
-	client_add(rv, process_cman, cluster_dead);
+	client_add(rv, process_cluster, cluster_dead);
+
+	update_cluster();
 
 	rv = setup_ccs();
 	if (rv < 0)
@@ -1184,10 +1193,10 @@ static void loop(void)
 		 * code in: cpg-new.c
 		 */
 
-		rv = setup_cpg();
+		rv = setup_cpg_daemon();
 		if (rv < 0)
 			goto out;
-		client_add(rv, process_cpg, cluster_dead);
+		client_add(rv, process_cpg_daemon, cluster_dead);
 
 		rv = set_protocol();
 		if (rv < 0)
@@ -1291,7 +1300,7 @@ static void loop(void)
 	}
  out:
 	if (group_mode == GROUP_LIBCPG)
-		close_cpg();
+		close_cpg_daemon();
 	else if (group_mode == GROUP_LIBGROUP) {
 		close_plocks();
 		close_cpg_old();
@@ -1300,7 +1309,7 @@ static void loop(void)
 		close_groupd();
 	close_logging();
 	close_ccs();
-	close_cman();
+	close_cluster();
 
 	if (!list_empty(&mountgroups))
 		log_error("mountgroups abandoned");
@@ -1565,7 +1574,6 @@ int poll_dlm;
 int plock_fd;
 int plock_ci;
 struct list_head mountgroups;
-int cman_quorate;
 int our_nodeid;
 char *clustername;
 char daemon_debug_buf[256];
