@@ -1034,6 +1034,8 @@ static void _receive_own(struct mountgroup *mg, char *buf, int len, int from)
 	int should_not_happen = 0;
 	int rv;
 
+	mg->last_plock_time = time(NULL);
+
 	memcpy(&info, buf + sizeof(struct gdlm_header), sizeof(info));
 	info_bswap_in(&info);
 
@@ -1200,6 +1202,8 @@ static void _receive_sync(struct mountgroup *mg, char *buf, int len, int from)
 	struct resource *r;
 	int rv;
 
+	mg->last_plock_time = time(NULL);
+
 	memcpy(&info, buf + sizeof(struct gdlm_header), sizeof(info));
 	info_bswap_in(&info);
 
@@ -1245,6 +1249,8 @@ static void _receive_drop(struct mountgroup *mg, char *buf, int len, int from)
 	struct resource *r;
 	int rv;
 
+	mg->last_plock_time = time(NULL);
+
 	memcpy(&info, buf + sizeof(struct gdlm_header), sizeof(info));
 	info_bswap_in(&info);
 
@@ -1492,7 +1498,7 @@ void process_plocks(int ci)
 	struct dlm_plock_info info;
 	struct timeval now;
 	uint64_t usec;
-	int rv;
+	int create, rv;
 
 	if (limit_plocks()) {
 		poll_ignore_plock = 1;
@@ -1547,7 +1553,9 @@ void process_plocks(int ci)
 		plock_rate_delays = 0;
 	}
 
-	rv = find_resource(mg, info.number, 1, &r);
+	create = (info.optype == DLM_PLOCK_OP_UNLOCK) ? 0 : 1;
+
+	rv = find_resource(mg, info.number, create, &r);
 	if (rv)
 		goto fail;
 
