@@ -1572,6 +1572,30 @@ dump_config_version(FILE *fp)
 }
 
 
+void
+dump_resource_info(FILE *fp)
+{
+	int x = central_events_enabled();
+
+	pthread_rwlock_rdlock(&resource_lock);
+	fprintf(fp, "=== Resource Agent Information ===\n");
+	dump_resource_rules(fp, &_rules);
+	fprintf(fp, "=== Defined Resources ===\n");
+	dump_resources(fp, &_resources);
+	fprintf(fp, "=== Failover Domains ===\n");
+	dump_domains(fp, &_domains);
+
+	if (x) {
+		fprintf(fp, "=== Events ===\n");
+		dump_events(fp, master_event_table);
+	}
+
+	fprintf(fp, "=== Resource Tree ===\n");
+	dump_resource_tree(fp, &_tree);
+	pthread_rwlock_unlock(&resource_lock);
+}
+
+
 /**
  Copy out the incarnations after doing CONDSTOPs
  */
@@ -1579,7 +1603,6 @@ static int
 copy_incarnations(resource_t **leftres, resource_t **rightres)
 {
 	resource_t *lc, *rc;
-	int ret;
 
 	list_do(leftres, lc) {
 		rc = find_resource_by_ref(rightres, lc->r_rule->rr_type,
@@ -1867,6 +1890,22 @@ check_restart(const char *rg_name)
 	return ret;
 }
 
+int
+clear_restart(const char *rg_name)
+{
+	resource_node_t *node;
+	int ret = 0;
+
+	pthread_rwlock_rdlock(&resource_lock);
+	node = node_by_ref(&_tree, rg_name);
+	if (node) {
+		ret = restart_clear(node->rn_restart_counter);
+	}
+	pthread_rwlock_unlock(&resource_lock);
+
+	return ret;
+}
+
 
 void
 kill_resource_groups(void)
