@@ -510,6 +510,8 @@ do_load_resource(int ccsfd, char *base,
 	char *ref;
 	resource_node_t *node;
 	resource_t *curres;
+	time_t failure_expire = 0;
+	int max_failures = 0;
 
 	snprintf(tok, sizeof(tok), "%s/@ref", base);
 
@@ -604,6 +606,36 @@ do_load_resource(int ccsfd, char *base,
 		free(ref);
 	}
 
+	/* per-resource-node failures / expire times */
+	snprintf(tok, sizeof(tok), "%s/@__max_failures", base);
+#ifndef NO_CCS
+	if (ccs_get(ccsfd, tok, &ref) == 0) {
+#else
+	if (conf_get(tok, &ref) == 0) {
+#endif
+		max_failures = atoi(ref);
+		if (max_failures < 0)
+			max_failures = 0;
+		free(ref);
+	}
+
+	snprintf(tok, sizeof(tok), "%s/@__failure_expire_time", base);
+#ifndef NO_CCS
+	if (ccs_get(ccsfd, tok, &ref) == 0) {
+#else
+	if (conf_get(tok, &ref) == 0) {
+#endif
+		failure_expire = (time_t)expand_time(ref);
+		if ((int64_t)failure_expire < 0)
+			failure_expire = 0;
+		free(ref);
+	}
+
+	if (max_failures && failure_expire) {
+		node->rn_failure_counter = restart_init(failure_expire,
+							max_failures);
+	}
+
 	curres->r_refs++;
 
 	*newnode = node;
@@ -851,6 +883,10 @@ destroy_resource_tree(resource_node_t **tree)
 			restart_cleanup(node->rn_restart_counter);
 		}
 
+		if (node->rn_failure_counter) {
+			restart_cleanup(node->rn_failure_counter);
+		}
+
 		if(node->rn_actions){
 			free(node->rn_actions);
 		}
@@ -860,48 +896,56 @@ destroy_resource_tree(resource_node_t **tree)
 
 
 static void
-_print_resource_tree(resource_node_t **tree, int level)
+_print_resource_tree(FILE *fp, resource_node_t **tree, int level)
 {
 	resource_node_t *node;
 	int x, y;
 
 	list_do(tree, node) {
 		for (x = 0; x < level; x++)
-			printf("  ");
+			fprintf(fp, "  ");
 
-		printf("%s", node->rn_resource->r_rule->rr_type);
+		fprintf(fp, "%s", node->rn_resource->r_rule->rr_type);
 		if (node->rn_flags) {
-			printf(" [ ");
+			fprintf(fp, " [ ");
+			if (node->rn_flags & RF_INLINE)
+				fprintf(fp, "INLINE ");
 			if (node->rn_flags & RF_NEEDSTOP)
-				printf("NEEDSTOP ");
+				fprintf(fp, "NEEDSTOP ");
 			if (node->rn_flags & RF_NEEDSTART)
-				printf("NEEDSTART ");
+				fprintf(fp, "NEEDSTART ");
 			if (node->rn_flags & RF_COMMON)
-				printf("COMMON ");
+				fprintf(fp, "COMMON ");
 			if (node->rn_flags & RF_INDEPENDENT)
-				printf("INDEPENDENT ");
+				fprintf(fp, "INDEPENDENT ");
+			if (node->rn_flags & RF_RECONFIG)
+				fprintf(fp, "RECONFIG ");
+			if (node->rn_flags & RF_INIT)
+				fprintf(fp, "INIT ");
+			if (node->rn_flags & RF_DESTROY)
+				fprintf(fp, "DESTROY ");
 			if (node->rn_flags & RF_ENFORCE_TIMEOUTS)
-				printf("ENFORCE-TIMEOUTS ");
-			printf("]");
+				fprintf(fp, "ENFORCE-TIMEOUTS ");
+			fprintf(fp, "]");
 		}
-		printf(" {\n");
+		fprintf(fp, " {\n");
 
 		for (x = 0; node->rn_resource->r_attrs &&
 		     node->rn_resource->r_attrs[x].ra_value; x++) {
 			for (y = 0; y < level+1; y++)
-				printf("  ");
-			printf("%s = \"%s\";\n",
-			       node->rn_resource->r_attrs[x].ra_name,
-			       attr_value(node,
-					  node->rn_resource->r_attrs[x].ra_name)
-			      );
+				fprintf(fp, "  ");
+			fprintf(fp, "%s = \"%s\";\n",
+				node->rn_resource->r_attrs[x].ra_name,
+ 				attr_value(node,
+					   node->rn_resource->r_attrs[x].ra_name)
+			       );
 		}
 
-		_print_resource_tree(&node->rn_child, level + 1);
+		_print_resource_tree(fp, &node->rn_child, level + 1);
 
 		for (x = 0; x < level; x++)
-			printf("  ");
-		printf("}\n");
+			fprintf(fp, "  ");
+		fprintf(fp, "}\n");
 	} while (!list_done(tree, node));
 }
 
@@ -909,7 +953,14 @@ _print_resource_tree(resource_node_t **tree, int level)
 void
 print_resource_tree(resource_node_t **tree)
 {
-	_print_resource_tree(tree, 0);
+	_print_resource_tree(stdout, tree, 0);
+}
+
+
+void
+dump_resource_tree(FILE *fp, resource_node_t **tree)
+{
+	_print_resource_tree(fp, tree, 0);
 }
 
 
@@ -1134,6 +1185,15 @@ do_status(resource_node_t *node)
 	 * completed. */
 	node->rn_actions[idx].ra_last = time(NULL);
 
+	/* If we have not exceeded our failure count threshold, then fudge
+	 * the status check this round */
+	if (x && node->rn_failure_counter) {
+		if (!restart_threshold_exceeded(node->rn_failure_counter)) {
+			x = 0;
+			restart_add(node->rn_failure_counter);
+		}
+	}
+
 	node->rn_last_status = x;
 	node->rn_last_depth = node->rn_actions[idx].ra_depth;
 	node->rn_checked = 1;
@@ -1195,9 +1255,8 @@ clear_checks(resource_node_t *node)
 {
 	time_t now;
 	int x = 0;
-	resource_t *res = node->rn_resource;
 
-	now = res->r_started;
+	now = get_time((char *)"start", 0, node);
 
 	for (; node->rn_actions[x].ra_name; x++) {
 
@@ -1208,6 +1267,8 @@ clear_checks(resource_node_t *node)
 		node->rn_actions[x].ra_last = now;
 	}
 
+	restart_clear(node->rn_failure_counter);
+
 	node->rn_checked = 0;
 	node->rn_last_status = 0;
 	node->rn_last_depth = 0;
@@ -1376,8 +1437,9 @@ _res_op_internal(resource_node_t __attribute__ ((unused)) **tree,
 		   incarnations there are. */
 		pthread_mutex_lock(&node->rn_resource->r_mutex);
 		if (node->rn_state == RES_STARTED) {
-			assert(node->rn_resource->r_incarnations > 0);
-			--node->rn_resource->r_incarnations;
+			assert(node->rn_resource->r_incarnations >= 0);
+			if (node->rn_resource->r_incarnations > 0)
+				--node->rn_resource->r_incarnations;
 		}
 
 		node->rn_flags &= ~RF_NEEDSTOP;
