@@ -580,6 +580,63 @@ out:
 }
 
 
+static int
+sl_migrate_service(void)
+{
+	char *svcname = NULL;
+	int target_node = 0;
+	int nargs, t, ret = -1;
+
+	nargs = SLang_Num_Function_Args;
+
+	/* Takes one, two, or three */
+	if (nargs != 2) {
+		SLang_verror(SL_Syntax_Error,
+		     (char *)"%s: Wrong # of args (%d), must be 2: service_name target_node\n",
+		     __FUNCTION__, nargs);
+		return -1;
+	}
+
+	t = SLang_peek_at_stack();
+	if (t != SLANG_INT_TYPE) {
+		SLang_verror(SL_Syntax_Error,
+			     (char *)"%s: expected type %d got %d\n",
+			     __FUNCTION__, SLANG_INT_TYPE, t);
+		goto out;
+	}
+
+	if (SLang_pop_integer(&target_node) < 0) {
+		SLang_verror(SL_Syntax_Error,
+		    (char *)"%s: Failed to pop integer from stack!\n",
+		    __FUNCTION__);
+		goto out;
+	}
+
+	t = SLang_peek_at_stack();
+	if (t != SLANG_STRING_TYPE) {
+		SLang_verror(SL_Syntax_Error,
+			     (char *)"%s: expected type %d got %d\n",
+			     __FUNCTION__,
+			     SLANG_STRING_TYPE, t);
+		goto out;
+	}
+
+	if (SLpop_string(&svcname) < 0) {
+		goto out;
+	}
+
+	ret = service_op_migrate(svcname, target_node);
+
+	if (ret == 0)
+		ret = target_node;
+out:
+	if (svcname)
+		free(svcname);
+	_user_return = ret;
+	return ret;
+}
+
+
 /* Take an array of integers given its length and
    push it on to the S/Lang stack */
 void
@@ -979,6 +1036,8 @@ static SLang_Intrin_Fun_Type rgmanager_slang[] =
 			 SLANG_INT_TYPE),
 	MAKE_INTRINSIC_0((char *)"service_start", sl_start_service,
 			 SLANG_INT_TYPE),
+	MAKE_INTRINSIC_0((char *)"service_migrate", sl_migrate_service,
+			 SLANG_INT_TYPE),
 	MAKE_INTRINSIC_S((char *)"service_status", sl_service_status,
 			 SLANG_VOID_TYPE),
 	MAKE_INTRINSIC_S((char *)"service_freeze", sl_service_freeze,
@@ -1103,22 +1162,38 @@ static int
 S_service_event(const char *file, const char *script, char *name,
 	        int state, int owner, int last_owner)
 {
+	char policy[20];
 	int ret;
 
 	_service_name = name;
 	_service_state = (char *)rg_state_str(state);
 	_service_owner = owner;
 	_service_last_owner = last_owner;
-	_service_restarts_exceeded = check_restart(name);
+	_service_restarts_exceeded = 0;
 
 	switch(state) {
 	case RG_STATE_DISABLED:
 	case RG_STATE_STOPPED:
 	case RG_STATE_FAILED:
-	case RG_STATE_RECOVER:
 	case RG_STATE_ERROR:
 		/* There is no owner for these states.  Ever.  */
 		_service_owner = -1;
+		break;
+	case RG_STATE_RECOVER:
+		get_recovery_policy(name, policy, sizeof(policy));
+		if (!strcasecmp(policy, "restart")) {
+			_service_restarts_exceeded = check_restart(name);
+			add_restart(name);
+		} else if (!strcasecmp(policy, "relocate")) {
+			/* Restart threshold is always exceeded 
+			 * with 'relocate' recovery policy */
+			_service_restarts_exceeded = 1;
+		}
+		_service_owner = -1;
+	}
+
+	if (_service_restarts_exceeded) {
+		clear_restart(name);
 	}
 
 	ret = do_slang_run(file, script);
@@ -1135,7 +1210,7 @@ S_service_event(const char *file, const char *script, char *name,
 
 static int
 S_user_event(const char *file, const char *script, char *name,
-	     int request, int arg1, int arg2, int target, msgctx_t *ctx)
+	     int request, int arg1, int arg2, int target, msgctx_t **ctx)
 {
 	int ret = RG_SUCCESS;
 
@@ -1159,20 +1234,21 @@ S_user_event(const char *file, const char *script, char *name,
 
 	/* XXX Send response code to caller - that 0 should be the
 	   new service owner, if there is one  */
-	if (ctx) {
+	if (*ctx) {
 		if (_user_return > 0) {
 			/* sl_start_service() squashes return code and
 			   node ID into one value.  <0 = error, >0 =
 			   success, return-value == node id running
 			   service */
-			send_ret(ctx, name, 0, request, _user_return);
+			send_ret(*ctx, name, 0, request, _user_return);
 		} else {
 			/* return value < 0 ... pass directly back;
 			   don't transpose */
-			send_ret(ctx, name, _user_return, request, 0);
+			send_ret(*ctx, name, _user_return, request, 0);
 		}
-		msg_close(ctx);
-		msg_free_ctx(ctx);
+		msg_close(*ctx);
+		msg_free_ctx(*ctx);
+		*ctx = NULL;
 	}
 	_user_return = 0;
 	return ret;
@@ -1213,7 +1289,7 @@ slang_do_script(event_t *pattern, event_t *ev)
 				ev->ev.user.u_arg1,
 				ev->ev.user.u_arg2,
 				ev->ev.user.u_target,
-				ev->ev.user.u_ctx);
+				&ev->ev.user.u_ctx);
 		break;
 	default:
 		break;
